/**
 * 
 */
package org.gcube.accounting.analytics.persistence.postgresql;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.SortedMap;
import java.util.SortedSet;

import org.gcube.accounting.analytics.Filter;
import org.gcube.accounting.analytics.Info;
import org.gcube.accounting.analytics.NumberedFilter;
import org.gcube.accounting.analytics.TemporalConstraint;
import org.gcube.accounting.analytics.TemporalConstraint.AggregationMode;
import org.gcube.accounting.analytics.exception.DuplicatedKeyFilterException;
import org.gcube.accounting.analytics.exception.KeyException;
import org.gcube.accounting.analytics.exception.ValueException;
import org.gcube.accounting.analytics.persistence.AccountingPersistenceBackendQueryConfiguration;
import org.gcube.accounting.analytics.persistence.AccountingPersistenceQuery;
import org.gcube.accounting.datamodel.UsageRecord;
import org.gcube.accounting.datamodel.aggregation.AggregatedServiceUsageRecord;
import org.gcube.accounting.utility.postgresql.RecordToDBMapping;
import org.gcube.documentstore.records.Record;
import org.junit.Before;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Luca Frosini (ISTI - CNR)
 */
public class AccountingPersistenceQueryPostgreSQLTest extends ContextTest {
	
	private static final Logger logger = LoggerFactory.getLogger(AccountingPersistenceQueryPostgreSQLTest.class);
	
	protected AccountingPersistenceQueryPostgreSQL accountingPersistenceQueryPostgreSQL;
	
	@Before
	public void before() throws Exception {
		AccountingPersistenceBackendQueryConfiguration configuration = new AccountingPersistenceBackendQueryConfiguration(
				AccountingPersistenceQueryPostgreSQL.class);
		
		accountingPersistenceQueryPostgreSQL = new AccountingPersistenceQueryPostgreSQL();
		accountingPersistenceQueryPostgreSQL.prepareConnection(configuration);
	}
	
	@Test
	public void testGetNoContextTimeSeries() throws DuplicatedKeyFilterException, KeyException, ValueException, Exception {
		Calendar startTimeCalendar = Calendar.getInstance();
		startTimeCalendar.set(Calendar.YEAR, 2020);
		startTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		startTimeCalendar.set(Calendar.DAY_OF_MONTH, 1);
		startTimeCalendar.set(Calendar.HOUR_OF_DAY, 0);
		startTimeCalendar.set(Calendar.MINUTE, 0);
		
		Calendar entTimeCalendar = Calendar.getInstance();
		entTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		entTimeCalendar.set(Calendar.DAY_OF_MONTH, 15);
		entTimeCalendar.set(Calendar.HOUR_OF_DAY, 16);
		entTimeCalendar.set(Calendar.MINUTE, 17);
		
		List<Filter> filters = new ArrayList<>();
		Filter filter = new Filter(UsageRecord.CONSUMER_ID, QueryTest.getRandomUser());
		filters.add(filter);
		
		TemporalConstraint temporalConstraint = new TemporalConstraint(startTimeCalendar.getTimeInMillis(), entTimeCalendar.getTimeInMillis(), AggregationMode.MINUTELY);
		SortedMap<Calendar, Info> timeseries = accountingPersistenceQueryPostgreSQL.getNoContextTimeSeries(AggregatedServiceUsageRecord.class, temporalConstraint, filters);
		for(Calendar c : timeseries.keySet()) {
			Info info = timeseries.get(c);
			logger.debug("{}", info);
		}
	}
	
	@Test
	public void testTimeSeries() throws DuplicatedKeyFilterException, KeyException, ValueException, Exception {
		Calendar startTimeCalendar = Calendar.getInstance();
		startTimeCalendar.set(Calendar.YEAR, 2020);
		startTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		startTimeCalendar.set(Calendar.DAY_OF_MONTH, 1);
		startTimeCalendar.set(Calendar.HOUR_OF_DAY, 0);
		startTimeCalendar.set(Calendar.MINUTE, 0);
		
		Calendar entTimeCalendar = Calendar.getInstance();
		entTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		entTimeCalendar.set(Calendar.DAY_OF_MONTH, 15);
		entTimeCalendar.set(Calendar.HOUR_OF_DAY, 16);
		entTimeCalendar.set(Calendar.MINUTE, 17);
		
		List<Filter> filters = new ArrayList<>();
		Filter filter = new Filter(UsageRecord.CONSUMER_ID, QueryTest.getRandomUser());
		filters.add(filter);
		
		TemporalConstraint temporalConstraint = new TemporalConstraint(startTimeCalendar.getTimeInMillis(), entTimeCalendar.getTimeInMillis(), AggregationMode.MINUTELY);
		SortedMap<Calendar, Info> timeseries = accountingPersistenceQueryPostgreSQL.getTimeSeries(AggregatedServiceUsageRecord.class, temporalConstraint, filters);
		for(Calendar c : timeseries.keySet()) {
			Info info = timeseries.get(c);
			logger.debug("{}", info);
		}
	}


	@Test
	public void testContextTimeSeries() throws DuplicatedKeyFilterException, KeyException, ValueException, Exception {
		Calendar startTimeCalendar = Calendar.getInstance();
		startTimeCalendar.set(Calendar.YEAR, 2020);
		startTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		startTimeCalendar.set(Calendar.DAY_OF_MONTH, 1);
		startTimeCalendar.set(Calendar.HOUR_OF_DAY, 0);
		startTimeCalendar.set(Calendar.MINUTE, 0);
		
		Calendar entTimeCalendar = Calendar.getInstance();
		entTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		entTimeCalendar.set(Calendar.DAY_OF_MONTH, 15);
		entTimeCalendar.set(Calendar.HOUR_OF_DAY, 16);
		entTimeCalendar.set(Calendar.MINUTE, 17);
		
		List<Filter> filters = new ArrayList<>();
		Filter filter = new Filter(UsageRecord.CONSUMER_ID, QueryTest.getRandomUser());
		filters.add(filter);
		
		TemporalConstraint temporalConstraint = new TemporalConstraint(startTimeCalendar.getTimeInMillis(), entTimeCalendar.getTimeInMillis(), AggregationMode.MONTHLY);
		
		List<String> contexts = new ArrayList<>();
		contexts.add("/gcube/devsec/devVRE");
		contexts.add("/gcube/devNext/NextNext");
		
		SortedMap<Filter, SortedMap<Calendar, Info>> contextTimeseries = accountingPersistenceQueryPostgreSQL.getContextTimeSeries(AggregatedServiceUsageRecord.class, temporalConstraint, filters, contexts);
		for(Filter f : contextTimeseries.keySet()) {
			logger.debug("{}", f);
			SortedMap<Calendar, Info> timeseries =  contextTimeseries.get(f);
			for(Calendar c : timeseries.keySet()) {
				Info info = timeseries.get(c);
				logger.debug("{}", info);
			}
		}
	}
	
	
	@Test
	public void testGetFilterValues() throws DuplicatedKeyFilterException, KeyException, ValueException, Exception {
		Calendar startTimeCalendar = Calendar.getInstance();
		startTimeCalendar.set(Calendar.YEAR, 2020);
		startTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		startTimeCalendar.set(Calendar.DAY_OF_MONTH, 1);
		startTimeCalendar.set(Calendar.HOUR_OF_DAY, 0);
		startTimeCalendar.set(Calendar.MINUTE, 0);
		
		Calendar entTimeCalendar = Calendar.getInstance();
		entTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		entTimeCalendar.set(Calendar.DAY_OF_MONTH, 15);
		entTimeCalendar.set(Calendar.HOUR_OF_DAY, 16);
		entTimeCalendar.set(Calendar.MINUTE, 17);
		
		List<Filter> filters = new ArrayList<>();
		Filter filter = new Filter(UsageRecord.CONSUMER_ID, QueryTest.getRandomUser());
		filters.add(filter);
		
		TemporalConstraint temporalConstraint = new TemporalConstraint(startTimeCalendar.getTimeInMillis(), entTimeCalendar.getTimeInMillis(), AggregationMode.MINUTELY);
		SortedSet<NumberedFilter> numberedFilters = accountingPersistenceQueryPostgreSQL.getFilterValues(AggregatedServiceUsageRecord.class, temporalConstraint, filters, AggregatedServiceUsageRecord.CALLED_METHOD);
		for(NumberedFilter numberedFilter  : numberedFilters) {
			logger.debug("{}", numberedFilter);
		}
		
		numberedFilters = accountingPersistenceQueryPostgreSQL.getFilterValues(AggregatedServiceUsageRecord.class, temporalConstraint, filters, AggregatedServiceUsageRecord.CALLED_METHOD);
		for(NumberedFilter numberedFilter  : numberedFilters) {
			logger.debug("{}", numberedFilter);
		}
		
		
		numberedFilters = accountingPersistenceQueryPostgreSQL.getFilterValues(AggregatedServiceUsageRecord.class, temporalConstraint, filters, AggregatedServiceUsageRecord.CALLED_METHOD, 3);
		for(NumberedFilter numberedFilter  : numberedFilters) {
			logger.debug("{}", numberedFilter);
		}
	}
	
	@Test
	public void testTopValues() throws DuplicatedKeyFilterException, KeyException, ValueException, Exception {	
		String orderingProperty = AccountingPersistenceQuery.getDefaultOrderingProperties(AggregatedServiceUsageRecord.class);
		
		Calendar startTimeCalendar = Calendar.getInstance();
		startTimeCalendar.set(Calendar.YEAR, 2020);
		startTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		startTimeCalendar.set(Calendar.DAY_OF_MONTH, 1);
		startTimeCalendar.set(Calendar.HOUR_OF_DAY, 0);
		startTimeCalendar.set(Calendar.MINUTE, 0);
		
		Calendar entTimeCalendar = Calendar.getInstance();
		entTimeCalendar.set(Calendar.MONTH, Calendar.MARCH);
		entTimeCalendar.set(Calendar.DAY_OF_MONTH, 15);
		entTimeCalendar.set(Calendar.HOUR_OF_DAY, 16);
		entTimeCalendar.set(Calendar.MINUTE, 17);
		
		List<Filter> filters = new ArrayList<>();
		Filter filter = new Filter(UsageRecord.CONSUMER_ID, QueryTest.getRandomUser());
		filters.add(filter);
		
		TemporalConstraint temporalConstraint = new TemporalConstraint(startTimeCalendar.getTimeInMillis(), entTimeCalendar.getTimeInMillis(), AggregationMode.MONTHLY);
		
		SortedMap<NumberedFilter, SortedMap<Calendar, Info>> topTimeSeries = accountingPersistenceQueryPostgreSQL.getTopValues(AggregatedServiceUsageRecord.class, temporalConstraint, filters, AggregatedServiceUsageRecord.CALLED_METHOD, orderingProperty);
		for(NumberedFilter numberedFilter : topTimeSeries.keySet()) {
			logger.debug("{}", numberedFilter);
			SortedMap<Calendar, Info> timeseries = topTimeSeries.get(numberedFilter);
			for(Calendar c : timeseries.keySet()) {
				Info info = timeseries.get(c);
				logger.debug("{}", info);
			}
		}		
	}
	
	@Test
	public void testGetRecord() throws Exception {
		String type = RecordToDBMapping.getRecordTypeByClass(AggregatedServiceUsageRecord.class);
		Record record = accountingPersistenceQueryPostgreSQL.getRecord("fa573711-ceb6-44ba-9c83-bd47e0915b80", type);
		logger.debug("{}", record);
	}
}